/*
    Copyright (C) 2015 Jolla Ltd.
    Copyright (C) 2018 Matti Lehtimäki <matti.lehtimaki@gmail.com>
    Contact: Aaron McCarthy <aaron.mccarthy@jollamobile.com>

    This file is part of geoclue-hybris.

    Geoclue-hybris is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License.
*/

#include "binderlocationbackend.h"

#include "hybrisprovider.h"

#include <QtNetwork/QHostAddress>

#include <strings.h>
#include <sys/time.h>

#define GNSS_BINDER_DEFAULT_DEV  "/dev/hwbinder"
#define GNSS_QTI_BINDER_DEFAULT_DEV  "/dev/vndbinder"

HybrisLocationBackend *getLocationBackend()
{
    return qobject_cast<HybrisLocationBackend *>(new BinderLocationBackend());
}

enum GnssFunctions {
    GNSS_SET_CALLBACK = 1,
    GNSS_START = 2,
    GNSS_STOP = 3,
    GNSS_CLEANUP = 4,
    GNSS_INJECT_TIME = 5,
    GNSS_INJECT_LOCATION = 6,
    GNSS_DELETE_AIDING_DATA = 7,
    GNSS_SET_POSITION_MODE = 8,
    GNSS_GET_EXTENSION_AGNSS_RIL = 9,
    GNSS_GET_EXTENSION_GNSS_GEOFENCING = 10,
    GNSS_GET_EXTENSION_AGNSS = 11,
    GNSS_GET_EXTENSION_GNSS_NI = 12,
    GNSS_GET_EXTENSION_GNSS_MEASUREMENT = 13,
    GNSS_GET_EXTENSION_GNSS_NAVIGATION_MESSAGE = 14,
    GNSS_GET_EXTENSION_XTRA = 15,
    GNSS_GET_EXTENSION_GNSS_CONFIGURATION = 16,
    GNSS_GET_EXTENSION_GNSS_DEBUG = 17,
    GNSS_GET_EXTENSION_GNSS_BATCHING = 18,
	//qti
	GNSS_GET_EXTENSION_HIDL_AGNSS = 19,
	GNSS_GET_EXTENSION_HIDL_DEBUGREPORTSERVICE = 20,
	GNSS_GET_EXTENSION_HIDL_FLPSERVICE = 21,
	GNSS_GET_EXTENSION_HIDL_GEOFENCESERVICE = 22,
	GNSS_GET_EXTENSION_HIDL_IZATNETWORKSERVICE = 23,
	GNSS_GET_EXTENSION_HIDL_IZATFUSEDSERVICE = 24,
	GNSS_GET_EXTENSION_HIDL_GNSSNISERVICE = 25,
	GNSS_GET_EXTENSION_HIDL_OSNPGLUESERVICE = 26,
	GNSS_GET_EXTENSION_HIDL_IZATSUBSCRIPTIONSERVICE = 27,
	GNSS_GET_EXTENSION_HIDL_RILINFOMONITORSERVICE = 28,
	GNSS_GET_EXTENSION_HIDL_XTSERVICE = 29,
};

enum GnssCallbacks {
    GNSS_LOCATION_CB = 1,
    GNSS_STATUS_CB = 2,
    GNSS_SV_STATUS_CB = 3,
    GNSS_NMEA_CB = 4,
    GNSS_SET_CAPABILITIES_CB = 5,
    GNSS_ACQUIRE_WAKELOCK_CB = 6,
    GNSS_RELEASE_WAKELOCK_CB = 7,
    GNSS_REQUEST_TIME_CB = 8,
    GNSS_SET_SYSTEM_INFO_CB = 9
};

enum GnssDebudFunctions {
    GNSS_DEBUG_GET_DEBUG_DATA = 1
};

enum GnssNiFunctions {
    GNSS_NI_SET_CALLBACK = 1,
    GNSS_NI_RESPOND = 2,
	GNSS_HIDLGNSSNI_SETVENDORCALLBACK = 3,
};

enum GnssNiCallbacks {
    GNSS_NI_NOTIFY_CB = 1
};

enum GnssXtraFunctions {
    GNSS_XTRA_SET_CALLBACK = 1,
    GNSS_XTRA_INJECT_XTRA_DATA = 2
};

enum GnssXtraCallbacks {
    GNSS_XTRA_DOWNLOAD_REQUEST_CB = 1
};

enum AGnssFunctions {
    AGNSS_SET_CALLBACK = 1,
    AGNSS_DATA_CONN_CLOSED = 2,
    AGNSS_DATA_CONN_FAILED = 3,
    AGNSS_SET_SERVER = 4,
    AGNSS_DATA_CONN_OPEN = 5
};

enum AGnssCallbacks {
    AGNSS_STATUS_IP_V4_CB = 1,
    AGNSS_STATUS_IP_V6_CB = 2
};

enum AGnssRilFunctions {
    AGNSS_RIL_SET_CALLBACK = 1,
    AGNSS_RIL_SET_REF_LOCATION = 2,
    AGNSS_RIL_SET_ID = 3,
    AGNSS_RIL_UPDATE_NETWORK_STATE = 4,
    AGNSS_RIL_UPDATE_NETWORK_AVAILABILITY = 5
};

enum AGnssRilCallbacks {
    AGNSS_RIL_REQUEST_REF_ID_CB = 1,
    AGNSS_RIL_REQUEST_REF_LOC_CB = 2
};

enum HybrisApnIpTypeEnum {
    HYBRIS_APN_IP_INVALID  = 0,
    HYBRIS_APN_IP_IPV4     = 1,
    HYBRIS_APN_IP_IPV6     = 2,
    HYBRIS_APN_IP_IPV4V6   = 3
};



#define GNSS_IFACE(x)       "android.hardware.gnss@1.0::" x
#define GNSS_REMOTE         GNSS_IFACE("IGnss")
#define GNSS_CALLBACK       GNSS_IFACE("IGnssCallback")
#define GNSS_DEBUG_REMOTE   GNSS_IFACE("IGnssDebug")
#define GNSS_NI_REMOTE      GNSS_IFACE("IGnssNi")
#define GNSS_NI_CALLBACK    GNSS_IFACE("IGnssNiCallback")
#define GNSS_XTRA_REMOTE    GNSS_IFACE("IGnssXtra")
#define GNSS_XTRA_CALLBACK  GNSS_IFACE("IGnssXtraCallback")
#define AGNSS_REMOTE        GNSS_IFACE("IAGnss")
#define AGNSS_CALLBACK      GNSS_IFACE("IAGnssCallback")
#define AGNSS_RIL_REMOTE    GNSS_IFACE("IAGnssRil")
#define AGNSS_RIL_CALLBACK  GNSS_IFACE("IAGnssRilCallback")


#define GNSS_IFACE_QTI(x)    "vendor.qti.gnss@1.0::" x
#define GNSS_REMOTE_QTI      GNSS_IFACE_QTI("ILocHidlGnss")
#define GNSS_REMOTE_QTI1     GNSS_IFACE("ILocHidlGnss")
#define GNSS_IZAT_REMOTE     GNSS_IFACE_QTI("ILocHidlIzatProvider")
#define GNSS_IZAT_CALLBACK   GNSS_IFACE_QTI("ILocHidlIzatProviderCallback")
#define GNSS_XT_REMOTE       GNSS_IFACE_QTI("ILocHidlXT")
#define GNSS_OSNPGLUE        GNSS_IFACE_QTI("ILocHidlIzatOsNpGlue")
#define GNSS_OSNPGLUE_CALLBACK              GNSS_IFACE_QTI("ILocHidlIzatOsNpGlueCallback")
#define GNSS_IZAT_SUBSCRIPTION              GNSS_IFACE_QTI("ILocHidlIzatSubscription")
#define GNSS_IZAT_SUBSCRIPTION_CALLBACK     GNSS_IFACE_QTI("ILocHidlIzatSubscriptionCallback")
#define GNSS_RILINFOMONITOR                 GNSS_IFACE_QTI("ILocHidlRilInfoMonitor")
#define GNSS_HIDL_GNSSNI                    GNSS_IFACE_QTI("ILocHidlGnssNi")
#define GNSS_HIDL_GNSSNI_CALLBACK           GNSS_IFACE_QTI("IGnssNiCallback")
#define GNSS_HIDL_AGNSS                     GNSS_IFACE_QTI("ILocHidlAGnss")
	
enum GnssIzatFunctions {
    GNSS_IZAT_INIT = 1,
    GNSS_IZAT_DEINIT = 2,
    GNSS_IZAT_ONENABLE = 3,
    GNSS_IZAT_ONDISABLE = 4,
    GNSS_IZAT_ONADDREQUEST = 5,
    GNSS_IZAT_ONREMOVEREQUEST = 6,
};
enum GnssIzatCallbacks {
	GNSS_IZAT_LOCATION_CHANGED_CB = 1,
	GNSS_IZAT_STATUS_CHANGED_CB = 2,
};
enum GnssXTFunctions {
    GNSS_XT_INIT = 1,
	GNSS_XT_SETUSERPREF = 2,	
};
enum GnssOSNpGlueFunctions {
    GNSS_OSNPGLUE_SETCALLBACK = 1,
	GNSS_OSNPGLUE_LOCATIONCHANGED = 2,	
};
enum GnssOsNpGlueCallbacks {
    GNSS_OSNPGLUE_ON_SETREQUEST = 1,			
	GNSS_OSNPGLUE_ON_STOPREQUEST = 2,
};

enum GnssIzatSubscriptionFunctions {
	GNSS_IZAT_SUBSCRIPTION_INIT = 1,
	GNSS_IZAT_SUBSCRIPTION_DEINIT = 2,
	GNSS_IZAT_SUBSCRIPTION_BDIUPDATE = 3,
	GNSS_IZAT_SUBSCRIPTION_SDIUPDATE = 4,
	GNSS_IZAT_SUBSCRIPTION_NIUPDATE = 5,
	GNSS_IZAT_SUBSCRIPTION_SIUPDATE = 6,
	GNSS_IZAT_SUBSCRIPTION_CELLLTEUPDATE = 7,
};
enum GnssIzatSubscriptionCallbacks {
    GNSS_IZAT_SUBSCRIPTION_REQUESTDATA = 1,
	GNSS_IZAT_SUBSCRIPTION_UPDATESUBSCRIBE = 2,
	GNSS_IZAT_SUBSCRIPTION_UNSUBSCRIBE = 3,
	GNSS_IZAT_SUBSCRIPTION_TURNONMODULE = 4,
	GNSS_IZAT_SUBSCRIPTION_TURNOFFMODULE = 5,
};
enum GnssRilInfoMonFunctions {
    GNSS_RILINFOMONITOR_INIT = 1,
};

enum GnssHidlGnssNiCallbacks {
	 GNSS_HIDLGNSSNI_ON_NINOTIFYCBEXT = 1,
	 GNSS_HIDLGNSSNI_GNSSCAPABILITIES = 2
};
enum GnssHidlAGnssCallbacks {
	 GNSS_HIDLAGNSS_SETCALLBACK = 1,
};
/*==========================================================================*
 * Implementation
 *==========================================================================*/
const int nmea_max = 10;
int       nmea_index = 0;
QStringList   nmeaL;

namespace
{

HybrisApnIpType fromContextProtocol(const QString &protocol)
{
    if (protocol == QLatin1String("ip"))
        return HYBRIS_APN_IP_IPV4;
    else if (protocol == QLatin1String("ipv6"))
        return HYBRIS_APN_IP_IPV6;
    else if (protocol == QLatin1String("dual"))
        return HYBRIS_APN_IP_IPV4V6;
    else
        return HYBRIS_APN_IP_INVALID;
}

const void *geoclue_binder_gnss_decode_struct1(
    GBinderReader *in,
    guint size)
{
    const void *result = Q_NULLPTR;
    GBinderBuffer *buf = gbinder_reader_read_buffer(in);

    if (buf && buf->size == size) {
        result = buf->data;
    }
    gbinder_buffer_free(buf);
    return result;
}

#define geoclue_binder_gnss_decode_struct(type,in) \
    ((const type*)geoclue_binder_gnss_decode_struct1(in, sizeof(type)))

bool nmeaChecksumValid(const QByteArray &nmea)
{
    unsigned char checksum = 0;
    for (int i = 1; i < nmea.length(); ++i) {
        if (nmea.at(i) == '*') {
            if (nmea.length() < i+3)
                return false;

            checksum ^= nmea.mid(i+1, 2).toInt(0, 16);

            break;
        }

        checksum ^= nmea.at(i);
    }

    return checksum == 0;
}

void parseRmc(const QByteArray &nmea)
{
    QList<QByteArray> fields = nmea.split(',');
    if (fields.count() < 12)
        return;

    bool ok;
    double variation = fields.at(10).toDouble(&ok);
    if (ok) {
        if (fields.at(11) == "W")
            variation = -variation;

        QMetaObject::invokeMethod(staticProvider, "setMagneticVariation", Qt::QueuedConnection,
                                  Q_ARG(double, variation));
    }
}

void processNmea(gint64 timestamp, const char *nmeaData)
{
    int length = strlen(nmeaData);
    while (length > 0 && isspace(nmeaData[length-1]))
        --length;

    if (length == 0)
        return;

    QByteArray nmea = QByteArray::fromRawData(nmeaData, length);

    qCDebug(lcGeoclueHybrisNmea) << timestamp << nmea;

    if (!nmeaChecksumValid(nmea))
        return;

    // truncate checksum and * from end of sentence
    nmea.truncate(nmea.length()-3);

    if (nmea.startsWith("$GPRMC"))
        parseRmc(nmea);
}

const double MpsToKnots = 1.943844;

GBinderLocalReply *geoclue_binder_gnss_callback(
    GBinderLocalObject *obj,
    GBinderRemoteRequest *req,
    guint code,
    guint flags,
    int *status,
    void *user_data)
{
    Q_UNUSED(flags)
    Q_UNUSED(user_data)
    const char *iface = gbinder_remote_request_interface(req);

    if (!g_strcmp0(iface, GNSS_CALLBACK)) {
        GBinderReader reader;

        gbinder_remote_request_init_reader(req, &reader);
		//qDebug() << "gnss_callback" << code;
        switch (code) {
        case GNSS_LOCATION_CB:
            {
            Location loc;

            const GnssLocation *location = geoclue_binder_gnss_decode_struct
                (GnssLocation, &reader);

            loc.setTimestamp(location->timestamp);

            if (location->gnssLocationFlags & HYBRIS_GNSS_LOCATION_HAS_LAT_LONG) {
                loc.setLatitude(location->latitudeDegrees);
                loc.setLongitude(location->longitudeDegrees);
            }

            if (location->gnssLocationFlags & HYBRIS_GNSS_LOCATION_HAS_ALTITUDE)
                loc.setAltitude(location->altitudeMeters);

            if (location->gnssLocationFlags & HYBRIS_GNSS_LOCATION_HAS_SPEED)
                loc.setSpeed(location->speedMetersPerSec * MpsToKnots);

            if (location->gnssLocationFlags & HYBRIS_GNSS_LOCATION_HAS_BEARING)
                loc.setDirection(location->bearingDegrees);

            if ((location->gnssLocationFlags & HYBRIS_GNSS_LOCATION_HAS_HORIZONTAL_ACCURACY) ||
                (location->gnssLocationFlags & HYBRIS_GNSS_LOCATION_HAS_VERTICAL_ACCURACY)) {
                Accuracy accuracy;
                if (location->gnssLocationFlags & HYBRIS_GNSS_LOCATION_HAS_HORIZONTAL_ACCURACY) {
                    accuracy.setHorizontal(location->horizontalAccuracyMeters);
                }
                if (location->gnssLocationFlags & HYBRIS_GNSS_LOCATION_HAS_VERTICAL_ACCURACY) {
                    accuracy.setVertical(location->verticalAccuracyMeters);
                }
                loc.setAccuracy(accuracy);
            }

            QMetaObject::invokeMethod(staticProvider, "setLocation", Qt::QueuedConnection,
                                      Q_ARG(Location, loc));
            }
            break;
        case GNSS_STATUS_CB:
            {
            guint32 stat;
            if (gbinder_reader_read_uint32(&reader, &stat)) {
				qDebug() << "GNSS_STATUS_CB stat " << stat;
                if (stat == HYBRIS_GNSS_STATUS_ENGINE_ON) {
                    QMetaObject::invokeMethod(staticProvider, "engineOn", Qt::QueuedConnection);
                }
                if (stat == HYBRIS_GNSS_STATUS_ENGINE_OFF) {
                    QMetaObject::invokeMethod(staticProvider, "engineOff", Qt::QueuedConnection);
                }
            }
            }
            break;
        case GNSS_SV_STATUS_CB:
            {
            const GnssSvStatus *svStatus = geoclue_binder_gnss_decode_struct
                (GnssSvStatus, &reader);

            QList<SatelliteInfo> satellites;
            QList<int> usedPrns;
		qCDebug(lcGeoclueHybris) <<"nb satellites" << svStatus->numSvs;
            for (int i = 0; i < svStatus->numSvs; ++i) {
                SatelliteInfo satInfo;
                GnssSvInfo svInfo = svStatus->gnssSvList[i];
                satInfo.setSnr(svInfo.cN0Dbhz);
                satInfo.setElevation(svInfo.elevationDegrees);
                satInfo.setAzimuth(svInfo.azimuthDegrees);
                int prn = svInfo.svid;
                // From https://github.com/barbeau/gpstest
                // and https://github.com/mvglasow/satstat/wiki/NMEA-IDs
                if (svInfo.constellation == GnssConstellationType::SBAS) {
                    prn -= 87;
                } else if (svInfo.constellation == GnssConstellationType::GLONASS) {
                    prn += 64;
                } else if (svInfo.constellation == GnssConstellationType::BEIDOU) {
                    prn += 200;
                } else if (svInfo.constellation == GnssConstellationType::GALILEO) {
                    prn += 300;
                }
                satInfo.setPrn(prn);
                satellites.append(satInfo);

                if (svInfo.svFlag & HYBRIS_GNSS_SV_FLAGS_USED_IN_FIX)
                    usedPrns.append(prn);
            }

            QMetaObject::invokeMethod(staticProvider, "setSatellite", Qt::QueuedConnection,
                                      Q_ARG(QList<SatelliteInfo>, satellites),
                                      Q_ARG(QList<int>, usedPrns));
            }
            break;
        case GNSS_NMEA_CB:
            {
            gint64 timestamp;
            if (gbinder_reader_read_int64(&reader, &timestamp)) {
                char *nmeaData = gbinder_reader_read_hidl_string(&reader);
                if (nmeaData) {
		    if (nmea_index >=nmea_max)
			nmea_index = 0;
		    if (nmeaL.size() <= nmea_index)
			nmeaL.push_back(nmeaData);
		    else
			nmeaL[nmea_index] = nmeaData;
		    nmea_index ++;
		    qCDebug(lcGeoclueHybris) <<"nmea" << nmeaData;
                    processNmea(timestamp, nmeaData);
                    g_free(nmeaData);
		    QMetaObject::invokeMethod(staticProvider, "setNmea", Qt::QueuedConnection,Q_ARG(QStringList, nmeaL));
                }
            }
            }
            break;
        case GNSS_SET_CAPABILITIES_CB:
            {
            guint32 capabilities;
            if (gbinder_reader_read_uint32(&reader, &capabilities)) {
                qDebug() << "capabilities" << showbase << hex << capabilities;
            }
            }
            break;
        case GNSS_ACQUIRE_WAKELOCK_CB:
        case GNSS_RELEASE_WAKELOCK_CB:
            break;
        case GNSS_REQUEST_TIME_CB:
            qDebug() << "GNSS request UTC time";
            QMetaObject::invokeMethod(staticProvider, "injectUtcTime", Qt::QueuedConnection);
            break;
        case GNSS_SET_SYSTEM_INFO_CB:
            qDebug() << "GNSS set system info";
            break;
        default:
            qWarning("Failed to decode callback %u", code);
            break;
        }
        *status = GBINDER_STATUS_OK;
        return gbinder_local_reply_append_int32(gbinder_local_object_new_reply(obj), 0);
    } else {
        qWarning("Unknown interface %s and code %u", iface, code);
        *status = GBINDER_STATUS_FAILED;
    }
    return Q_NULLPTR;
}

GBinderLocalReply *geoclue_binder_gnss_xtra_callback(
    GBinderLocalObject *obj,
    GBinderRemoteRequest *req,
    guint code,
    guint flags,
    int *status,
    void *user_data)
{
    Q_UNUSED(flags)
    Q_UNUSED(user_data)
    const char *iface = gbinder_remote_request_interface(req);

    if (!g_strcmp0(iface, GNSS_XTRA_CALLBACK)) {
        GBinderReader reader;

        gbinder_remote_request_init_reader(req, &reader);
        switch (code) {
        case GNSS_XTRA_DOWNLOAD_REQUEST_CB:
            QMetaObject::invokeMethod(staticProvider, "xtraDownloadRequest", Qt::QueuedConnection);
            break;
        default:
            qWarning("Failed to decode callback %u", code);
            break;
        }
        *status = GBINDER_STATUS_OK;
        return gbinder_local_reply_append_int32(gbinder_local_object_new_reply(obj), 0);
    } else {
        qWarning("Unknown interface %s and code %u", iface, code);
        *status = GBINDER_STATUS_FAILED;
    }
    return Q_NULLPTR;
}

GBinderLocalReply *geoclue_binder_agnss_callback(
    GBinderLocalObject *obj,
    GBinderRemoteRequest *req,
    guint code,
    guint flags,
    int *status,
    void *user_data)
{
    Q_UNUSED(flags)
    Q_UNUSED(user_data)
    const char *iface = gbinder_remote_request_interface(req);

    if (!g_strcmp0(iface, AGNSS_CALLBACK)) {
        GBinderReader reader;

        gbinder_remote_request_init_reader(req, &reader);
        switch (code) {
	qDebug() <<"agnss_callback" << code;
        case AGNSS_STATUS_IP_V4_CB:
            {
            QHostAddress ipv4;
            QHostAddress ipv6;
            QByteArray ssid;
            QByteArray password;

            const AGnssStatusIpV4 *status = geoclue_binder_gnss_decode_struct
                (AGnssStatusIpV4, &reader);

            ipv4.setAddress(status->ipV4Addr);

            QMetaObject::invokeMethod(staticProvider, "agpsStatus", Qt::QueuedConnection,
                                      Q_ARG(qint16, status->type), Q_ARG(quint16, status->status),
                                      Q_ARG(QHostAddress, ipv4), Q_ARG(QHostAddress, ipv6),
                                      Q_ARG(QByteArray, ssid), Q_ARG(QByteArray, password));
            }
            break;
        case AGNSS_STATUS_IP_V6_CB:
            {
            QHostAddress ipv4;
            QHostAddress ipv6;
            QByteArray ssid;
            QByteArray password;

            const AGnssStatusIpV6 *status = geoclue_binder_gnss_decode_struct
                (AGnssStatusIpV6, &reader);

            ipv6.setAddress(status->ipV6Addr);

            QMetaObject::invokeMethod(staticProvider, "agpsStatus", Qt::QueuedConnection,
                                      Q_ARG(qint16, status->type), Q_ARG(quint16, status->status),
                                      Q_ARG(QHostAddress, ipv4), Q_ARG(QHostAddress, ipv6),
                                      Q_ARG(QByteArray, ssid), Q_ARG(QByteArray, password));
            }
            break;
        default:
            qWarning("Failed to decode callback %u", code);
            break;
        }
        *status = GBINDER_STATUS_OK;
        return gbinder_local_reply_append_int32(gbinder_local_object_new_reply(obj), 0);
    } else {
        qWarning("Unknown interface %s and code %u", iface, code);
        *status = GBINDER_STATUS_FAILED;
    }
    return Q_NULLPTR;
}


GBinderLocalReply *geoclue_binder_agnss_ril_callback(
    GBinderLocalObject *obj,
    GBinderRemoteRequest *req,
    guint code,
    guint flags,
    int *status,
    void *user_data)
{
    Q_UNUSED(flags)
    Q_UNUSED(user_data)
    const char *iface = gbinder_remote_request_interface(req);

    if (!g_strcmp0(iface, AGNSS_RIL_CALLBACK)) {
        GBinderReader reader;

        gbinder_remote_request_init_reader(req, &reader);
        switch (code) {
        case AGNSS_RIL_REQUEST_REF_ID_CB:
            qCDebug(lcGeoclueHybris) << "AGNSS RIL request ref ID";
            break;
        case AGNSS_RIL_REQUEST_REF_LOC_CB:
            qCDebug(lcGeoclueHybris) << "AGNSS RIL request ref location";
            break;
        default:
            qWarning("Failed to decode callback %u", code);
            break;
        }
        *status = GBINDER_STATUS_OK;
        return gbinder_local_reply_append_int32(gbinder_local_object_new_reply(obj), 0);
    } else {
        qWarning("Unknown interface %s and code %u", iface, code);
        *status = GBINDER_STATUS_FAILED;
    }
    return Q_NULLPTR;
}


GBinderLocalReply *geoclue_binder_gnss_ni_callback(
    GBinderLocalObject *obj,
    GBinderRemoteRequest *req,
    guint code,
    guint flags,
    int *status,
    void *user_data)
{
    Q_UNUSED(flags)
    Q_UNUSED(user_data)
    const char *iface = gbinder_remote_request_interface(req);

    if (!g_strcmp0(iface, GNSS_NI_CALLBACK)) {
        GBinderReader reader;

        gbinder_remote_request_init_reader(req, &reader);
        switch (code) {
        case GNSS_NI_NOTIFY_CB:
            qCDebug(lcGeoclueHybris) << "GNSS NI notify";
            break;
        default:
            qWarning("Failed to decode callback %u", code);
            break;
        }
        *status = GBINDER_STATUS_OK;
        return gbinder_local_reply_append_int32(gbinder_local_object_new_reply(obj), 0);
    } else {
        qWarning("Unknown interface %s and code %u", iface, code);
        *status = GBINDER_STATUS_FAILED;
    }
    return Q_NULLPTR;
}

GBinderLocalReply *geoclue_binder_gnss_izat_callback(
    GBinderLocalObject *obj,
    GBinderRemoteRequest *req,
    guint code,
    guint flags,
    int *status,
    void *user_data)
{
    Q_UNUSED(flags)
    Q_UNUSED(user_data)
    const char *iface = gbinder_remote_request_interface(req);

    if (!g_strcmp0(iface, GNSS_IZAT_CALLBACK)) {
        GBinderReader reader;

        gbinder_remote_request_init_reader(req, &reader);
        switch (code) {
        case GNSS_IZAT_LOCATION_CHANGED_CB: //LocHidlGnssLocation
			{
				const LocHidlGnssLocation *location = geoclue_binder_gnss_decode_struct
					(LocHidlGnssLocation, &reader);

				qDebug() << "Izat location" << location->latitude << location->longitude;				
				break;
			}
		case GNSS_IZAT_STATUS_CHANGED_CB:
			{
				qDebug() << "Izat status";
				break;
			}
        default:
			{
				qWarning("Failed to decode callback %u", code);
				break;
			}
        }
        *status = GBINDER_STATUS_OK;
        return gbinder_local_reply_append_int32(gbinder_local_object_new_reply(obj), 0);
    } else {
        qWarning("Unknown interface %s and code %u", iface, code);
        *status = GBINDER_STATUS_FAILED;
    }
    return Q_NULLPTR;
}
GBinderLocalReply *geoclue_binder_gnss_osnpglue_callback(
    GBinderLocalObject *obj,
    GBinderRemoteRequest *req,
    guint code,
    guint flags,
    int *status,
    void *user_data)
{
    Q_UNUSED(flags)
    Q_UNUSED(user_data)
    const char *iface = gbinder_remote_request_interface(req);

    if (!g_strcmp0(iface, GNSS_OSNPGLUE_CALLBACK)) {
        GBinderReader reader;

        gbinder_remote_request_init_reader(req, &reader);
        switch (code) {
        case GNSS_OSNPGLUE_ON_SETREQUEST: //LocHidlGnssLocation
			{
				qDebug() << "osnpglue onsetrequest";				
				break;
			}
		case GNSS_OSNPGLUE_ON_STOPREQUEST:
			{
				qDebug() << "osnpglue onstoprequest";
				break;
			}
        default:
			{
				qWarning("Failed to decode callback %u", code);
				break;
			}
        }
        *status = GBINDER_STATUS_OK;
        return gbinder_local_reply_append_int32(gbinder_local_object_new_reply(obj), 0);
    } else {
        qWarning("Unknown interface %s and code %u", iface, code);
        *status = GBINDER_STATUS_FAILED;
    }
    return Q_NULLPTR;
}

void geoclue_binder_gnss_gnss_died(
    GBinderRemoteObject */*obj*/,
    void *user_data)
{
    BinderLocationBackend *self = (BinderLocationBackend *)user_data;
    self->dropGnss();
}
GBinderLocalReply *geoclue_binder_gnss_izatsub_callback(
    GBinderLocalObject *obj,
    GBinderRemoteRequest *req,
    guint code,
    guint flags,
    int *status,
    void *user_data)
{
    Q_UNUSED(flags)
    Q_UNUSED(user_data)
    const char *iface = gbinder_remote_request_interface(req);

    if (!g_strcmp0(iface, GNSS_IZAT_SUBSCRIPTION_CALLBACK)) {
        GBinderReader reader;

        gbinder_remote_request_init_reader(req, &reader);
        switch (code) {
        case GNSS_IZAT_SUBSCRIPTION_REQUESTDATA:
			{
				
				qDebug() << "Izat subscription requestdata" ;				
				break;
			}
		case GNSS_IZAT_SUBSCRIPTION_UPDATESUBSCRIBE:
			{
				qDebug() << "Izat subscription updatesubscribe" ;
				break;
			}
		case GNSS_IZAT_SUBSCRIPTION_UNSUBSCRIBE:
			{
				qDebug() << "Izat subscription unsubscribe" ;
				break;
			}
		case GNSS_IZAT_SUBSCRIPTION_TURNONMODULE:
			{
				qDebug() << "Izat subscription turnonmodule" ;
				break;
			}
		case GNSS_IZAT_SUBSCRIPTION_TURNOFFMODULE:
			{
				qDebug() << "Izat subscription turnoffmodule" ;
				break;
			}
        default:
			{
				qWarning("Failed to decode callback %u", code);
				break;
			}
        }
        *status = GBINDER_STATUS_OK;
        return gbinder_local_reply_append_int32(gbinder_local_object_new_reply(obj), 0);
    } else {
        qWarning("Unknown interface %s and code %u", iface, code);
        *status = GBINDER_STATUS_FAILED;
    }
    return Q_NULLPTR;
}
//
GBinderLocalReply *geoclue_binder_gnss_hidlgnssni_callback(
    GBinderLocalObject *obj,
    GBinderRemoteRequest *req,
    guint code,
    guint flags,
    int *status,
    void *user_data)
{
    Q_UNUSED(flags)
    Q_UNUSED(user_data)
    const char *iface = gbinder_remote_request_interface(req);

    if (!g_strcmp0(iface, GNSS_IZAT_SUBSCRIPTION_CALLBACK)) {
        GBinderReader reader;

        gbinder_remote_request_init_reader(req, &reader);
        switch (code) {
        case GNSS_HIDLGNSSNI_ON_NINOTIFYCBEXT:
			{
				
				qDebug() << "hidl gnss ni ninotifycbext" ;				
				break;
			}
		case GNSS_HIDLGNSSNI_GNSSCAPABILITIES:
			{
				
				qDebug() << "hidl gnss ni gnsscapabilities" ;				
				break;
			}
        default:
			{
				qWarning("Failed to decode callback %u", code);
				break;
			}
        }
        *status = GBINDER_STATUS_OK;
        return gbinder_local_reply_append_int32(gbinder_local_object_new_reply(obj), 0);
    } else {
        qWarning("Unknown interface %s and code %u", iface, code);
        *status = GBINDER_STATUS_FAILED;
    }
    return Q_NULLPTR;
}
}

/*==========================================================================*
 * Backend class
 *==========================================================================*/

BinderLocationBackend::BinderLocationBackend(QObject *parent)
:   HybrisLocationBackend(parent), m_death_id(0), m_fqname(Q_NULLPTR), m_sm(Q_NULLPTR),
    m_clientGnss(Q_NULLPTR), m_remoteGnss(Q_NULLPTR), m_callbackGnss(Q_NULLPTR),
    m_clientGnssDebug(Q_NULLPTR), m_remoteGnssDebug(Q_NULLPTR),
    m_clientGnssNi(Q_NULLPTR), m_remoteGnssNi(Q_NULLPTR), m_callbackGnssNi(Q_NULLPTR),
    m_clientGnssXtra(Q_NULLPTR), m_remoteGnssXtra(Q_NULLPTR), m_callbackGnssXtra(Q_NULLPTR),
    m_clientAGnss(Q_NULLPTR), m_remoteAGnss(Q_NULLPTR), m_callbackAGnss(Q_NULLPTR),
    m_clientAGnssRil(Q_NULLPTR), m_remoteAGnssRil(Q_NULLPTR), m_callbackAGnssRil(Q_NULLPTR),
	 m_fqname_qti(Q_NULLPTR), m_sm_qti(Q_NULLPTR),m_clientGnssQti(Q_NULLPTR), m_remoteGnssQti(Q_NULLPTR)
{
}

BinderLocationBackend::~BinderLocationBackend()
{
    dropGnss();
}

void BinderLocationBackend::dropGnss()
{
    if (m_callbackGnss) {
        gbinder_local_object_drop(m_callbackGnss);
        m_callbackGnss = Q_NULLPTR;
    }
    if (m_clientGnss) {
        gbinder_client_unref(m_clientGnss);
        m_clientGnss = Q_NULLPTR;
    }
    if (m_remoteGnss) {
        gbinder_remote_object_remove_handler(m_remoteGnss, m_death_id);
        gbinder_remote_object_unref(m_remoteGnss);
        m_death_id = 0;
        m_remoteGnss = Q_NULLPTR;
    }
    if (m_clientGnssDebug) {
        gbinder_client_unref(m_clientGnssDebug);
        m_clientGnssDebug = Q_NULLPTR;
    }
    if (m_remoteGnssDebug) {
        gbinder_remote_object_unref(m_remoteGnssDebug);
        m_remoteGnssDebug = Q_NULLPTR;
    }
    if (m_callbackGnssNi) {
        gbinder_local_object_drop(m_callbackGnssNi);
        m_callbackGnssNi = Q_NULLPTR;
    }
    if (m_clientGnssNi) {
        gbinder_client_unref(m_clientGnssNi);
        m_clientGnssNi = Q_NULLPTR;
    }
    if (m_remoteGnssNi) {
        gbinder_remote_object_unref(m_remoteGnssNi);
        m_remoteGnssNi = Q_NULLPTR;
    }
    if (m_callbackGnssXtra) {
        gbinder_local_object_drop(m_callbackGnssXtra);
        m_callbackGnssXtra = Q_NULLPTR;
    }
    if (m_clientGnssXtra) {
        gbinder_client_unref(m_clientGnssXtra);
        m_clientGnssXtra = Q_NULLPTR;
    }
    if (m_remoteGnssXtra) {
        gbinder_remote_object_unref(m_remoteGnssXtra);
        m_remoteGnssXtra = Q_NULLPTR;
    }
    if (m_callbackAGnss) {
        gbinder_local_object_drop(m_callbackAGnss);
        m_callbackAGnss = Q_NULLPTR;
    }
    if (m_clientAGnss) {
        gbinder_client_unref(m_clientAGnss);
        m_clientAGnss = Q_NULLPTR;
    }
    if (m_remoteAGnss) {
        gbinder_remote_object_unref(m_remoteAGnss);
        m_remoteAGnss = Q_NULLPTR;
    }
    if (m_callbackAGnssRil) {
        gbinder_local_object_drop(m_callbackAGnssRil);
        m_callbackAGnssRil = Q_NULLPTR;
    }
    if (m_clientAGnssRil) {
        gbinder_client_unref(m_clientAGnssRil);
        m_clientAGnssRil = Q_NULLPTR;
    }
    if (m_remoteAGnssRil) {
        gbinder_remote_object_unref(m_remoteAGnssRil);
        m_remoteAGnssRil = Q_NULLPTR;
    }
    if (m_sm) {
        gbinder_servicemanager_unref(m_sm);
        m_sm = Q_NULLPTR;
    }

    g_free(m_fqname);
    m_fqname = Q_NULLPTR;
}

bool BinderLocationBackend::isReplySuccess(GBinderRemoteReply *reply)
{
    GBinderReader reader;
    gint32 status;
    gboolean result;
    gbinder_remote_reply_init_reader(reply, &reader);

    if (!gbinder_reader_read_int32(&reader, &status) || status != 0) {
        return false;
    }
    if (!gbinder_reader_read_bool(&reader, &result) || !result) {
        return false;
    }

    return true;
}

GBinderRemoteObject *BinderLocationBackend::getExtensionObject(GBinderRemoteReply *reply)
{
    GBinderReader reader;
    gint32 status;
    gbinder_remote_reply_init_reader(reply, &reader);

    if (!gbinder_reader_read_int32(&reader, &status) || status != 0) {
        qWarning("Failed to get extension object %d", status);
        return Q_NULLPTR;
    }

    return gbinder_reader_read_object(&reader);
}

// Gnss
bool BinderLocationBackend::gnssInit()
{
    bool ret = false;

    qWarning("Initialising GNSS interface");

    m_sm = gbinder_servicemanager_new(GNSS_BINDER_DEFAULT_DEV);
    if (m_sm) {
        int status = 0;

        /* Fetch remote reference from hwservicemanager */
        m_fqname = g_strconcat(GNSS_REMOTE "/default", Q_NULLPTR);
        m_remoteGnss = gbinder_servicemanager_get_service_sync(m_sm,
            m_fqname, &status);

        if (m_remoteGnss) {
            GBinderLocalRequest *req;
            GBinderRemoteReply *reply;

            /* get_service returns auto-released reference,
             * we need to add a reference of our own */
            gbinder_remote_object_ref(m_remoteGnss);
            m_clientGnss = gbinder_client_new(m_remoteGnss, GNSS_REMOTE);
            m_death_id = gbinder_remote_object_add_death_handler
                (m_remoteGnss, geoclue_binder_gnss_gnss_died, this);
            m_callbackGnss = gbinder_servicemanager_new_local_object
                (m_sm, GNSS_CALLBACK, geoclue_binder_gnss_callback, this);

            /* IGnss::setCallback */
			
            req = gbinder_client_new_request(m_clientGnss);
            gbinder_local_request_append_local_object(req, m_callbackGnss);
            reply = gbinder_client_transact_sync_reply(m_clientGnss,
                GNSS_SET_CALLBACK, req, &status);

            if (!status) {
                ret = isReplySuccess(reply);
            }

            gbinder_local_request_unref(req);
            gbinder_remote_reply_unref(reply);
			
        }
    }

    if (!ret) {
        qWarning("Failed to initialise GNSS interface");
    }
    return ret;
}

bool BinderLocationBackend::gnssStart()
{
    bool ret = false;
    if (m_clientGnss) {
        int status = 0;
        GBinderRemoteReply *reply;

		qWarning("Starting Positioning");
        reply = gbinder_client_transact_sync_reply(m_clientGnss,
            GNSS_START, Q_NULLPTR, &status);

        if (!status) {
            ret = isReplySuccess(reply);
        }

        gbinder_remote_reply_unref(reply);
    }

    if (!ret) {
        qWarning("Failed to start positioning");
    }
    return ret;
}

bool BinderLocationBackend::gnssStop()
{
    bool ret = false;

    if (m_clientGnss) {
        int status = 0;
        GBinderRemoteReply *reply;

        reply = gbinder_client_transact_sync_reply(m_clientGnss,
            GNSS_STOP, Q_NULLPTR, &status);

        if (!status) {
            ret = isReplySuccess(reply);
        }

        gbinder_remote_reply_unref(reply);
    }

    if (!ret) {
        qWarning("Failed to stop positioning");
    }
    return ret;
}

void BinderLocationBackend::gnssCleanup()
{
    if (m_clientGnss) {
        gbinder_client_transact(m_clientGnss, GNSS_CLEANUP, 0, NULL, NULL, NULL, NULL);
    }
}

bool BinderLocationBackend::gnssInjectLocation(double latitudeDegrees, double longitudeDegrees, float accuracyMeters)
{
    bool ret = false;

    if (m_clientGnss) {
        int status = 0;

        GBinderLocalRequest *req;
        GBinderRemoteReply *reply;
        GBinderWriter writer;

        req = gbinder_client_new_request(m_clientGnss);
        gbinder_local_request_init_writer(req, &writer);
        gbinder_writer_append_double(&writer, latitudeDegrees);
        gbinder_writer_append_double(&writer, longitudeDegrees);
        gbinder_writer_append_float(&writer, accuracyMeters);
        reply = gbinder_client_transact_sync_reply(m_clientGnss,
            GNSS_INJECT_LOCATION, req, &status);

        if (!status) {
            ret = isReplySuccess(reply);
        }
        if (!ret) {
            qWarning("Failed to inject location");
        }

        gbinder_local_request_unref(req);
        gbinder_remote_reply_unref(reply);
    }
    return ret;
}

bool BinderLocationBackend::gnssInjectTime(HybrisGnssUtcTime timeMs, int64_t timeReferenceMs, int32_t uncertaintyMs)
{
    bool ret = false;

    if (m_clientGnss) {
        int status = 0;
        GBinderLocalRequest *req;
        GBinderRemoteReply *reply;
        GBinderWriter writer;

        req = gbinder_client_new_request(m_clientGnss);
        gbinder_local_request_init_writer(req, &writer);
        gbinder_writer_append_int64(&writer, timeMs);
        gbinder_writer_append_int64(&writer, timeReferenceMs);
        gbinder_writer_append_int32(&writer, uncertaintyMs);

        reply = gbinder_client_transact_sync_reply(m_clientGnss,
            GNSS_INJECT_TIME, req, &status);

        if (!status) {
            ret = isReplySuccess(reply);
        }

        if (!ret) {
            qWarning("Failed to inject time");
        }
        gbinder_local_request_unref(req);
        gbinder_remote_reply_unref(reply);
    }
    return ret;
}

void BinderLocationBackend::gnssDeleteAidingData(HybrisGnssAidingData aidingDataFlags)
{
    if (m_clientGnss) {
        GBinderLocalRequest *req;

        req = gbinder_client_new_request(m_clientGnss);
        gbinder_local_request_append_int32(req, aidingDataFlags);
        gbinder_client_transact(m_clientGnss, GNSS_DELETE_AIDING_DATA, 
                                0, req, NULL, NULL, NULL);

        gbinder_local_request_unref(req);
    }
}

bool BinderLocationBackend::gnssSetPositionMode(HybrisGnssPositionMode mode, HybrisGnssPositionRecurrence recurrence,
                                       uint32_t minIntervalMs, uint32_t preferredAccuracyMeters,
                                       uint32_t preferredTimeMs)
{
    bool ret = false;

    if (m_clientGnss) {
        int status = 0;
        GBinderLocalRequest *req;
        GBinderRemoteReply *reply;
        GBinderWriter writer;

        req = gbinder_client_new_request(m_clientGnss);
        gbinder_local_request_init_writer(req, &writer);
        gbinder_writer_append_int32(&writer, mode);
        gbinder_writer_append_int32(&writer, recurrence);
        gbinder_writer_append_int32(&writer, minIntervalMs);
        gbinder_writer_append_int32(&writer, preferredAccuracyMeters);
        gbinder_writer_append_int32(&writer, preferredTimeMs);
        reply = gbinder_client_transact_sync_reply(m_clientGnss,
            GNSS_SET_POSITION_MODE, req, &status);

        if (!status) {
            ret = isReplySuccess(reply);
        }

        if (!ret) {
            qWarning("GNSS set position mode failed");
        }
        gbinder_local_request_unref(req);
        gbinder_remote_reply_unref(reply);
    }
    return ret;
}

// GnssDebug
void BinderLocationBackend::gnssDebugInit()
{
    GBinderRemoteReply *reply;
    int status = 0;

    reply = gbinder_client_transact_sync_reply(m_clientGnss,
        GNSS_GET_EXTENSION_GNSS_DEBUG, Q_NULLPTR, &status);

    if (!status) {
        m_remoteGnssDebug = getExtensionObject(reply);
        if (m_remoteGnssDebug) {
            qWarning("Initialising GNSS Debug interface");
            m_clientGnssDebug = gbinder_client_new(m_remoteGnssDebug, GNSS_DEBUG_REMOTE);
        }
    }
    gbinder_remote_reply_unref(reply);
}

// GnnNi
void BinderLocationBackend::gnssNiInit()
{
    GBinderRemoteReply *reply;
    int status = 0;

    reply = gbinder_client_transact_sync_reply(m_clientGnss,
        GNSS_GET_EXTENSION_GNSS_NI, Q_NULLPTR, &status);

    if (!status) {
        m_remoteGnssNi = getExtensionObject(reply);

        if (m_remoteGnssNi) {
            qWarning("Initialising GNSS NI interface");
            GBinderLocalRequest *req;
            m_clientGnssNi = gbinder_client_new(m_remoteGnssNi, GNSS_NI_REMOTE);
            m_callbackGnssNi = gbinder_servicemanager_new_local_object
                (m_sm, GNSS_NI_CALLBACK, geoclue_binder_gnss_ni_callback, this);

            gbinder_remote_reply_unref(reply);

            /* IGnssNi::setCallback */
            req = gbinder_client_new_request(m_clientGnssNi);
            gbinder_local_request_append_local_object(req, m_callbackGnssNi);
            reply = gbinder_client_transact_sync_reply(m_clientGnssNi,
                GNSS_NI_SET_CALLBACK, req, &status);

            if (!status) {
                if (!gbinder_remote_reply_read_int32(reply, &status) || status != 0) {
                    qWarning("Initialising GNSS NI interface failed %d", status);
                }
            }
            gbinder_local_request_unref(req);
        }
    }
    gbinder_remote_reply_unref(reply);
}

void BinderLocationBackend::gnssNiRespond(int32_t notifId, HybrisGnssUserResponseType userResponse)
{
    if (m_clientGnssNi) {
        int status = 0;
        GBinderLocalRequest *req;
        GBinderRemoteReply *reply;
        GBinderWriter writer;

        req = gbinder_client_new_request(m_clientGnssNi);
        gbinder_local_request_init_writer(req, &writer);
        gbinder_writer_append_int32(&writer, notifId);
        gbinder_writer_append_int32(&writer, userResponse);

        reply = gbinder_client_transact_sync_reply(m_clientGnssNi,
            GNSS_NI_RESPOND, req, &status);

        if (!status) {
            if (!gbinder_remote_reply_read_int32(reply, &status) || status != 0) {
                qWarning("GNSS NI respond failed %d", status);
            }
        }

        gbinder_local_request_unref(req);
        gbinder_remote_reply_unref(reply);
    }
}

// GnssXtra
void BinderLocationBackend::gnssXtraInit()
{
    GBinderRemoteReply *reply;
    int status = 0;

    reply = gbinder_client_transact_sync_reply(m_clientGnss,
        GNSS_GET_EXTENSION_XTRA, Q_NULLPTR, &status);

    if (!status) {
        m_remoteGnssXtra = getExtensionObject(reply);
		qDebug() << "gnssXtraInit m_remoteGnssXtra" << m_remoteGnssXtra;
        if (m_remoteGnssXtra) {
            qWarning("Initialising GNSS Xtra interface");
            GBinderLocalRequest *req;
            m_clientGnssXtra = gbinder_client_new(m_remoteGnssXtra, GNSS_XTRA_REMOTE);
            m_callbackGnssXtra = gbinder_servicemanager_new_local_object
                (m_sm, GNSS_XTRA_CALLBACK, geoclue_binder_gnss_xtra_callback, this);

            gbinder_remote_reply_unref(reply);

            /* IGnssXtra::setCallback */
            req = gbinder_client_new_request(m_clientGnssXtra);
            gbinder_local_request_append_local_object(req, m_callbackGnssXtra);
            reply = gbinder_client_transact_sync_reply(m_clientGnssXtra,
                GNSS_XTRA_SET_CALLBACK, req, &status);

            if (status || !isReplySuccess(reply)) {
                qWarning("Initialising GNSS Xtra interface failed");
            }
            gbinder_local_request_unref(req);
        }
    }
    gbinder_remote_reply_unref(reply);
}

bool BinderLocationBackend::gnssXtraInjectXtraData(QByteArray &xtraData)
{
    bool ret = false;
    if (m_clientGnssXtra) {
        int status = 0;

        GBinderLocalRequest *req;
        GBinderRemoteReply *reply;

        req = gbinder_client_new_request(m_clientGnssXtra);
        gbinder_local_request_append_hidl_string(req, xtraData.constData());
        reply = gbinder_client_transact_sync_reply(m_clientGnssXtra,
            GNSS_XTRA_INJECT_XTRA_DATA, req, &status);

        if (!status) {
            ret = isReplySuccess(reply);
        }

        if (!ret) {
            qWarning("GNSS Xtra inject xtra data failed");
        }
        gbinder_local_request_unref(req);
        gbinder_remote_reply_unref(reply);
    }
    return ret;
}

// AGnss
void BinderLocationBackend::aGnssInit()
{
    GBinderRemoteReply *reply;
    int status = 0;

    
    reply = gbinder_client_transact_sync_reply(m_clientGnss,
											   GNSS_GET_EXTENSION_AGNSS, Q_NULLPTR, &status);

    if (!status) {
        m_remoteAGnss = getExtensionObject(reply);

        if (m_remoteAGnss) {
            qWarning("Initialising AGNSS interface");
            GBinderLocalRequest *req;
            m_clientAGnss = gbinder_client_new(m_remoteAGnss, AGNSS_REMOTE);
            m_callbackAGnss = gbinder_servicemanager_new_local_object
                (m_sm, AGNSS_CALLBACK, geoclue_binder_agnss_callback, this);

            gbinder_remote_reply_unref(reply);

            /* IAGnss::setCallback */
            req = gbinder_client_new_request(m_clientAGnss);
            gbinder_local_request_append_local_object(req, m_callbackAGnss);
            reply = gbinder_client_transact_sync_reply(m_clientAGnss,
                AGNSS_SET_CALLBACK, req, &status);

            if (!status) {
                if (!gbinder_remote_reply_read_int32(reply, &status) || status != 0) {
                    qWarning("Initialising AGNSS interface failed %d", status);
                }
            }
            gbinder_local_request_unref(req);
        }
    }
	else {
		qWarning("Initialising AGNSS interface 1  failed %d", status);
	}
    gbinder_remote_reply_unref(reply);
}

bool BinderLocationBackend::aGnssDataConnClosed()
{
    int status = 0;
    bool ret = false;
    GBinderRemoteReply *reply;

    reply = gbinder_client_transact_sync_reply(m_clientAGnss,
        AGNSS_DATA_CONN_CLOSED, Q_NULLPTR, &status);

    if (!status) {
        ret = isReplySuccess(reply);
    }

    gbinder_remote_reply_unref(reply);
    return ret;
}

bool BinderLocationBackend::aGnssDataConnFailed()
{
    int status = 0;
    bool ret = false;
    GBinderRemoteReply *reply;

    reply = gbinder_client_transact_sync_reply(m_clientAGnss,
        AGNSS_DATA_CONN_FAILED, Q_NULLPTR, &status);

    if (!status) {
        ret = isReplySuccess(reply);
    }

    gbinder_remote_reply_unref(reply);
    return ret;
}

bool BinderLocationBackend::aGnssDataConnOpen(const QByteArray &apn, const QString &protocol)
{
    int status = 0;
    bool ret = false;
    GBinderLocalRequest *req;
    GBinderRemoteReply *reply;
    GBinderWriter writer;

    req = gbinder_client_new_request(m_clientAGnss);

    gbinder_local_request_init_writer(req, &writer);
    gbinder_writer_append_hidl_string(&writer, apn.constData());
    gbinder_writer_append_int32(&writer, fromContextProtocol(protocol));
    reply = gbinder_client_transact_sync_reply(m_clientAGnss,
        AGNSS_DATA_CONN_OPEN, req, &status);

    if (!status) {
        ret = isReplySuccess(reply);
    }

    gbinder_local_request_unref(req);
    gbinder_remote_reply_unref(reply);

    return ret;
}

int BinderLocationBackend::aGnssSetServer(HybrisAGnssType type, const char* hostname, int port)
{
    int status = 0;
    bool ret = false;
    GBinderLocalRequest *req;
    GBinderRemoteReply *reply;
    GBinderWriter writer;

    req = gbinder_client_new_request(m_clientAGnss);

    qWarning() <<"aGnssSetServer" << hostname << port;
    gbinder_local_request_init_writer(req, &writer);
    gbinder_writer_append_int32(&writer, type);
    gbinder_writer_append_hidl_string(&writer, hostname);
    gbinder_writer_append_int32(&writer, port);
    reply = gbinder_client_transact_sync_reply(m_clientAGnss,
        AGNSS_SET_SERVER, req, &status);

    if (!status) {
        ret = isReplySuccess(reply);
    }

    gbinder_local_request_unref(req);
    gbinder_remote_reply_unref(reply);

    return ret;
}

// AGnssRil
void BinderLocationBackend::aGnssRilInit()
{
    GBinderRemoteReply *reply;
    int status = 0;

    reply = gbinder_client_transact_sync_reply(m_clientGnss,
        GNSS_GET_EXTENSION_AGNSS_RIL, Q_NULLPTR, &status);

    if (!status) {
        m_remoteAGnssRil = getExtensionObject(reply);

        if (m_remoteAGnssRil) {
            qWarning("Initialising AGNSS RIL interface");
            GBinderLocalRequest *req;
            m_clientAGnssRil = gbinder_client_new(m_remoteAGnssRil, AGNSS_RIL_REMOTE);
            m_callbackAGnssRil = gbinder_servicemanager_new_local_object
                (m_sm, AGNSS_RIL_CALLBACK, geoclue_binder_agnss_ril_callback, this);

            gbinder_remote_reply_unref(reply);

            /* IAGnssRil::setCallback */
            req = gbinder_client_new_request(m_clientAGnssRil);
            gbinder_local_request_append_local_object(req, m_callbackAGnssRil);
            reply = gbinder_client_transact_sync_reply(m_clientAGnssRil,
                AGNSS_RIL_SET_CALLBACK, req, &status);

            if (!status) {
                if (!gbinder_remote_reply_read_int32(reply, &status) || status != 0) {
                    qWarning("Initialising AGNSS RIL interface failed %d", status);
                }
            }
            gbinder_local_request_unref(req);
        }
    }
    gbinder_remote_reply_unref(reply);
}
bool BinderLocationBackend::gnssInitQti()
{
    bool ret = false;

    qWarning("Initialising GNSS QTI interface");

    m_sm_qti = gbinder_servicemanager_new(GNSS_BINDER_DEFAULT_DEV);
    if (m_sm_qti) {
        int status = 0;

        /* Fetch remote reference from hwservicemanager */
        m_fqname_qti = g_strconcat(GNSS_REMOTE_QTI "/gnss_vendor", Q_NULLPTR);
        m_remoteGnssQti = gbinder_servicemanager_get_service_sync(m_sm_qti,
            m_fqname_qti, &status);
        if (m_remoteGnssQti) {
            GBinderLocalRequest *req;
            GBinderRemoteReply *reply;

            /* get_service returns auto-released reference,
             * we need to add a reference of our own */
            gbinder_remote_object_ref(m_remoteGnssQti);
            m_clientGnssQti = gbinder_client_new(m_remoteGnssQti, GNSS_REMOTE_QTI);
			m_death_id_qti = gbinder_remote_object_add_death_handler
                (m_remoteGnssQti, geoclue_binder_gnss_gnss_died, this);
			req = gbinder_client_new_request(m_clientGnssQti);
			ret = true;
        }
    }
	
    if (!ret) {
        qWarning("Failed to initialise GNSS QTI interface");
    }
    return ret;
}

void BinderLocationBackend::gnssIzatInit()
{
    GBinderRemoteReply *reply;
    int status = 0;

    reply = gbinder_client_transact_sync_reply(m_clientGnssQti,
		   GNSS_GET_EXTENSION_HIDL_IZATNETWORKSERVICE, Q_NULLPTR, &status);
    if (!status) {
        m_remoteGnssIzat = getExtensionObject(reply);

        if (m_remoteGnssIzat) {
            qWarning("Initialising GNSS QTI Izat interface");
            GBinderLocalRequest *req;
			
            m_clientGnssIzat = gbinder_client_new(m_remoteGnssIzat, GNSS_IZAT_REMOTE);
            m_callbackGnssIzat = gbinder_servicemanager_new_local_object
                (m_sm_qti, GNSS_IZAT_CALLBACK, geoclue_binder_gnss_izat_callback, this);

            gbinder_remote_reply_unref(reply);

            /* IGnssXtra::setCallback */
            req = gbinder_client_new_request(m_clientGnssIzat);
            gbinder_local_request_append_local_object(req, m_callbackGnssIzat);
            reply = gbinder_client_transact_sync_reply(m_clientGnssIzat,
                GNSS_IZAT_INIT, req, &status);

            if (status || !isReplySuccess(reply)) {
                qWarning("Initialising GNSS Izat interface failed");
            }
            gbinder_local_request_unref(req);
        }
	else {
		qWarning("Failed to Init GNSS QTI Izat interface 2");
	}
    }
    else
	qWarning("Failed to Init GNSS QTI Izat interface 1");
    gbinder_remote_reply_unref(reply);
}
bool BinderLocationBackend::gnssIzatOnEnable()
{
    bool ret = false;
    if (m_clientGnssIzat) {
        int status = 0;

        GBinderLocalRequest *req;
        GBinderRemoteReply *reply;

        req = gbinder_client_new_request(m_clientGnssIzat);        
        reply = gbinder_client_transact_sync_reply(m_clientGnssIzat,
            GNSS_IZAT_ONENABLE, req, &status);

        if (!status) {
            ret = isReplySuccess(reply);
        }

        if (!ret) {
            qWarning("GNSS IZAT OnEnable failed");
        }
	else
		qWarning("GNSS IZAT OnEnable ok");
        gbinder_local_request_unref(req);
        gbinder_remote_reply_unref(reply);
    }
    return ret;
}
bool BinderLocationBackend::gnssIzatOnAddRequest()
{
    bool ret = false;
    if (m_clientGnssIzat) {
        int status = 0;

        GBinderLocalRequest *req;
        GBinderRemoteReply *reply;
		GBinderWriter writer;
		LocHidlIzatRequest r;
		
        req = gbinder_client_new_request(m_clientGnssIzat);
		
		gbinder_local_request_init_writer(req, &writer);
	
		memset(&r,0,sizeof(r));
		r.provider = LocHidlIzatStreamType::NETWORK;
		r.numUpdates = 0x05;
		r.suggestedResponseTimeForFirstFix = 0x100;
		r.timeIntervalBetweenFixes = 0x30;
		r.smallestDistanceBetweenFixes = 0x30;
		r.suggestedHorizontalAccuracy = LocHidlIzatHorizontalAccuracy::BLOCK;
		r.suggestedAltitudeAccuracy = LocHidlIzatOtherAccuracy::LOW;
		r.suggestedBearingAccuracy = LocHidlIzatOtherAccuracy::LOW;
		gbinder_writer_append_buffer_object(&writer,(void *)&r,sizeof(r));
		
		//gbinder_local_request_append_local_object(req, (GBinderLocalObject*)&r);
        reply = gbinder_client_transact_sync_reply(m_clientGnssIzat,
												   GNSS_IZAT_ONADDREQUEST, req, &status);

        if (!status) {
            ret = isReplySuccess(reply);
        }

        if (!ret) {
            qWarning("GNSS IZAT OnAddRequest failed");
        }
		else
			qWarning("GNSS IZAT OnAddRequest ok");
        gbinder_local_request_unref(req);
        gbinder_remote_reply_unref(reply);
    }
    return ret;
}
void BinderLocationBackend::gnssXTInit()
{
    GBinderRemoteReply *reply;
    int status = 0;

    reply = gbinder_client_transact_sync_reply(m_clientGnssQti,
		   GNSS_GET_EXTENSION_HIDL_XTSERVICE, Q_NULLPTR, &status);
    if (!status) {
		m_remoteGnssXT = getExtensionObject(reply);
        if (m_remoteGnssXT) {
            qWarning("Initialising GNSS QTI XT interface");
            GBinderLocalRequest *req;

            m_clientGnssXT = gbinder_client_new(m_remoteGnssXT, GNSS_XT_REMOTE);

			qDebug() << "XT" << m_clientGnssXT;
            gbinder_remote_reply_unref(reply);

            req = gbinder_client_new_request(m_clientGnssXT);
            reply = gbinder_client_transact_sync_reply(m_clientGnssXT,
                GNSS_XT_INIT, req, &status);
            if (status) {
                qWarning("Initialising GNSS XT interface failed");
            }
            gbinder_local_request_unref(req);
        }
	else {
		qWarning("Failed to Init GNSS QTI XT interface 2");
	}
    }
    else
	qWarning("Failed to Init GNSS QTI XT interface 1");
    gbinder_remote_reply_unref(reply);
	gnssXTSetUserPref();
}
void BinderLocationBackend::gnssXTSetUserPref()
{
    bool ret = false;
	
    if (m_clientGnssXT) {
        int status = 0;

        GBinderLocalRequest *req;
        GBinderRemoteReply *reply;
		GBinderWriter writer;
		gboolean val = 1;
		
        req = gbinder_client_new_request(m_clientGnssXT);
		gbinder_local_request_init_writer(req, &writer);
		gbinder_writer_append_bool(&writer, val);
        reply = gbinder_client_transact_sync_reply(m_clientGnssXT,
            GNSS_XT_SETUSERPREF, req, &status);        
        if (status) {
            qWarning("GNSS XT SetUserPref failed");
        }
		else
			qWarning("GNSS XT SetUserPref ok");
		gbinder_local_request_unref(req);
    }
    return ;
}
void BinderLocationBackend::gnssOSNPGlue()
{
    GBinderRemoteReply *reply;
    int status = 0;

    reply = gbinder_client_transact_sync_reply(m_clientGnssQti,
		   GNSS_GET_EXTENSION_HIDL_OSNPGLUESERVICE, Q_NULLPTR, &status);
    if (!status) {
		m_remoteGnssOsNpGlue = getExtensionObject(reply);
        if (m_remoteGnssOsNpGlue) {
            qWarning("Initialising GNSS QTI OSNPGLUE interface");
            GBinderLocalRequest *req;

            m_clientGnssOsNpGlue = gbinder_client_new(m_remoteGnssOsNpGlue, GNSS_OSNPGLUE);
			m_callbackGnssOsNpGlue = gbinder_servicemanager_new_local_object
                (m_sm_qti, GNSS_OSNPGLUE_CALLBACK, geoclue_binder_gnss_osnpglue_callback, this);
            gbinder_remote_reply_unref(reply);

            req = gbinder_client_new_request(m_clientGnssOsNpGlue);
            gbinder_local_request_append_local_object(req, m_callbackGnssOsNpGlue);
            reply = gbinder_client_transact_sync_reply(m_clientGnssOsNpGlue,
                GNSS_OSNPGLUE_SETCALLBACK, req, &status);
            if (status) {
                qWarning("Initialising GNSS OSNPGLUE interface failed");
            }
            gbinder_local_request_unref(req);
        }
	else {
		qWarning("Failed to Init GNSS QTI OSNPGLUE interface 2");
	}
    }
    else
	qWarning("Failed to Init GNSS QTI OSNPGLUE interface 1");
    gbinder_remote_reply_unref(reply);
}
void BinderLocationBackend::gnssIzatSubInit()
{
    GBinderRemoteReply *reply;
    int status = 0;

    reply = gbinder_client_transact_sync_reply(m_clientGnssQti,
		   GNSS_GET_EXTENSION_HIDL_IZATSUBSCRIPTIONSERVICE, Q_NULLPTR, &status);
    if (!status) {
        m_remoteGnssIzatSub = getExtensionObject(reply);

        if (m_remoteGnssIzatSub) {
            qWarning("Initialising GNSS QTI Izat Subscription interface");
            GBinderLocalRequest *req;
			
            m_clientGnssIzatSub = gbinder_client_new(m_remoteGnssIzatSub, GNSS_IZAT_SUBSCRIPTION);
            m_callbackGnssIzatSub = gbinder_servicemanager_new_local_object
                (m_sm_qti, GNSS_IZAT_SUBSCRIPTION_CALLBACK, geoclue_binder_gnss_izatsub_callback, this);

            gbinder_remote_reply_unref(reply);

            req = gbinder_client_new_request(m_clientGnssIzatSub);
            gbinder_local_request_append_local_object(req, m_callbackGnssIzatSub);
            reply = gbinder_client_transact_sync_reply(m_clientGnssIzatSub,
                GNSS_IZAT_SUBSCRIPTION_INIT, req, &status);

            if (status || !isReplySuccess(reply)) {
                qWarning() << "Initialising GNSS Izat Subscription interface failed ";
            }
            gbinder_local_request_unref(req);
        }
	else {
		qWarning("Failed to Init GNSS QTI Izat Subscription interface 2");
	}
    }
    else
	qWarning("Failed to Init GNSS QTI Izat Subscription interface 1");
    gbinder_remote_reply_unref(reply);
}
void BinderLocationBackend::gnssIzatRilnfoMonInit()
{
    GBinderRemoteReply *reply;
    int status = 0;

    reply = gbinder_client_transact_sync_reply(m_clientGnssQti,
		   GNSS_GET_EXTENSION_HIDL_RILINFOMONITORSERVICE, Q_NULLPTR, &status);
    if (!status) {
        m_remoteGnssRilInfoMon = getExtensionObject(reply);

        if (m_remoteGnssRilInfoMon) {
            qWarning("Initialising GNSS QTI Ril Info Mon interface");
            GBinderLocalRequest *req;
			
            m_clientGnssRilInfoMon = gbinder_client_new(m_remoteGnssRilInfoMon, GNSS_RILINFOMONITOR);
          
            gbinder_remote_reply_unref(reply);

            req = gbinder_client_new_request(m_clientGnssRilInfoMon);
            reply = gbinder_client_transact_sync_reply(m_clientGnssRilInfoMon,
                GNSS_RILINFOMONITOR_INIT, req, &status);

            if (status) {
                qWarning() << "Initialising GNSS  RilInfoMon interface failed ";
            }
            gbinder_local_request_unref(req);
        }
	else {
		qWarning("Failed to Init GNSS QTI  RilInfoMon interface 2");
	}
    }
    else
	qWarning("Failed to Init GNSS QTI  RilInfoMon interface 1");
    gbinder_remote_reply_unref(reply);
}
void BinderLocationBackend::gnsshidlgnssniinit()
{
    GBinderRemoteReply *reply;
    int status = 0;

    reply = gbinder_client_transact_sync_reply(m_clientGnssQti,
		   GNSS_GET_EXTENSION_HIDL_GNSSNISERVICE, Q_NULLPTR, &status);
    if (!status) {
        m_remoteGnssHidlGnssNi = getExtensionObject(reply);

        if (m_remoteGnssHidlGnssNi) {
            qWarning("Initialising GNSS QTI GNSS NI interface");
            GBinderLocalRequest *req;
			
            m_clientGnssHidlGnssNi = gbinder_client_new(m_remoteGnssHidlGnssNi, GNSS_HIDL_GNSSNI);
            m_callbackGnssHidlGnssNi = gbinder_servicemanager_new_local_object
                (m_sm_qti, GNSS_HIDL_GNSSNI_CALLBACK , geoclue_binder_gnss_hidlgnssni_callback, this);

            gbinder_remote_reply_unref(reply);

            req = gbinder_client_new_request(m_clientGnssHidlGnssNi);
            gbinder_local_request_append_local_object(req, m_callbackGnssHidlGnssNi);
            reply = gbinder_client_transact_sync_reply(m_clientGnssHidlGnssNi,
                GNSS_HIDLGNSSNI_SETVENDORCALLBACK, req, &status);

            if (status) {
                qWarning() << "Initialising GNSS QTI GNSS NI interface failed ";
            }
            gbinder_local_request_unref(req);
        }
	else {
		qWarning("Failed to Init GNSS QTI GNSS NI interface 2");
	}
    }
    else
	qWarning("Failed to Init GNSS QTI GNSS NI interface 1");
    gbinder_remote_reply_unref(reply);
}
void BinderLocationBackend::gnsshidlagnssinit()
{
    GBinderRemoteReply *reply;
    int status = 0;

    reply = gbinder_client_transact_sync_reply(m_clientGnssQti,
			   GNSS_GET_EXTENSION_HIDL_AGNSS , Q_NULLPTR, &status);
    if (!status) {
        m_remoteGnssHidlAGnss = getExtensionObject(reply);

        if (m_remoteGnssHidlAGnss) {
            qWarning("Initialising GNSS QTI AGNSS  interface");
            GBinderLocalRequest *req;
			
            m_clientGnssHidlAGnss = gbinder_client_new(m_remoteGnssHidlAGnss, AGNSS_REMOTE /*GNSS_HIDL_AGNSS*/);
            m_callbackGnssHidlAGnss = gbinder_servicemanager_new_local_object
                (m_sm_qti, AGNSS_CALLBACK, geoclue_binder_agnss_callback, this);

            gbinder_remote_reply_unref(reply);

            req = gbinder_client_new_request(m_clientGnssHidlAGnss);
            gbinder_local_request_append_local_object(req, m_callbackGnssHidlAGnss);
            reply = gbinder_client_transact_sync_reply(m_clientGnssHidlAGnss,
                AGNSS_SET_CALLBACK, req, &status);

            if (status) {
                qWarning() << "Initialising GNSS QTI AGNSS  interface failed ";
            }
            gbinder_local_request_unref(req);
        }
	else {
		qWarning("Failed to Init GNSS QTI AGNSS  interface 2");
	}
    }
    else
	qWarning("Failed to Init GNSS QTI AGNSS  interface 1");
    gbinder_remote_reply_unref(reply);
}
